import pytest
import tempfile
import os
import yaml
import flaky
import io as sio
from jsonschema.exceptions import ValidationError
from yggdrasil import yamlfile
from yaml.constructor import ConstructorError
from tests import TestBase as base_class
_yaml_env = 'TEST_YAML_FILE'


def direct_translate(msg):  # pragma: no cover
    r"""Test translator that just returns passed message."""
    return msg


def test_load_yaml():
    r"""Test loading yaml."""
    cwd = os.getcwd()
    fname = os.path.join(cwd, 'test_load_yaml.yml')
    dict_write = {'test': 'hello'}
    dict_read = {'test': 'hello',
                 'working_dir': cwd}
    contents = yaml.dump(dict_write)
    with open(fname, 'w') as fd:
        yaml.dump(dict_write, fd)
    try:
        # Dictionary
        out = yamlfile.load_yaml(dict_write)
        assert(out == dict_read)
        # File name
        out = yamlfile.load_yaml(fname)
        assert(out == dict_read)
        # Open file object
        with open(fname, 'r') as fd:
            out = yamlfile.load_yaml(fd)
            assert(out == dict_read)
        # Open stream
        out = yamlfile.load_yaml(sio.StringIO(contents))
        assert(out == dict_read)
    finally:
        # Remove file
        if os.path.isfile(fname):
            os.remove(fname)


def test_load_yaml_error():
    r"""Test error on loading invalid file."""
    with pytest.raises(IOError):
        yamlfile.load_yaml('invalid')


def test_parse_component_error():
    r"""Test errors in parse_component."""
    with pytest.raises(yamlfile.YAMLSpecificationError):
        yamlfile.parse_component(1, 'invalid', 'invalid')
    with pytest.raises(yamlfile.YAMLSpecificationError):
        yamlfile.parse_component({}, 'invalid', 'invalid')


@flaky.flaky(max_runs=3)
def test_load_yaml_git():
    r"""Test loading a yaml from a remote git repository."""
    import git
    yml = "https://github.com/cropsinsilico/example-fakemodel/fakemodel.yml"
    with pytest.raises(Exception):
        yamlfile.load_yaml(yml)
    assert('model' in yamlfile.load_yaml('git:' + yml))
    yml = "cropsinsilico/example-fakemodel/fakemodel.yml"
    assert('model' in yamlfile.load_yaml('git:' + yml))
    git.rmtree("cropsinsilico")
    

class YamlTestBase(base_class):
    r"""Test base for yamlfile."""
    
    _contents = tuple()
    _include = tuple()
    _use_json = False
    _parse_kwargs = dict()

    @pytest.fixture(scope="class")
    def contents(self):
        r"""tuple: Contents of files."""
        return self._contents

    @pytest.fixture(scope="class")
    def nfiles(self, contents):
        r"""int: Number of files created for the test."""
        return len(contents)

    @pytest.fixture(scope="class")
    def use_json(self):
        r"""bool: Whether or not to use JSON for the test."""
        return self._use_json

    @pytest.fixture(scope="class")
    def parse_kwargs(self):
        r"""dict: Keyword arguments for the parse call."""
        return self._parse_kwargs

    @pytest.fixture(scope="class")
    def include(self):
        r"""tuple: Contents of included files."""
        return self._include

    @pytest.fixture(scope="class")
    def yaml_env(self):
        r"""str: Environment variable where file path is stored."""
        return _yaml_env

    @pytest.fixture
    def files(self, nfiles, get_fname):
        r"""list: Name of test files."""
        return [get_fname(i) for i in range(nfiles)]

    @pytest.fixture
    def include_files(self, include, get_fname):
        r"""list: Names of test include files."""
        return [get_fname(i, '_incl') for i in range(len(include))]

    @pytest.fixture(autouse=True)
    def create_files(self, files, contents, yaml_env,
                     include_files, include, nfiles):
        r"""Write contents to temp file."""
        if nfiles > 0:
            os.environ[yaml_env] = files[0]
        for fname, content in zip(files, contents):
            with open(fname, 'w') as f:
                f.write('\n'.join(content))
        for i, (fname, content) in enumerate(zip(include_files,
                                                 include)):
            os.environ['%s%d' % (yaml_env, i)] = os.path.join(
                '.', os.path.basename(fname))
            with open(fname, 'w') as f:
                f.write('\n'.join(content))
        yield
        for fname in files + include_files:
            if os.path.isfile(fname):
                os.remove(fname)

    @pytest.fixture
    def get_fname(self, uuid, use_json):
        r"""Path to temporary file."""
        def wrapped_get_fname(idx=0, suffix=''):
            if use_json:
                ext = '.json'
            else:
                ext = '.yml'
            return os.path.join(tempfile.gettempdir(),
                                '%s_%s_%d%s%s' % (
                                    tempfile.gettempprefix(), uuid,
                                    idx, suffix, ext))
        return wrapped_get_fname

    def test_parse_yaml(self, parse_kwargs, nfiles, files):
        r"""Test successfully reading & parsing yaml."""
        if nfiles == 0:
            pass
        elif nfiles == 1:
            yamlfile.parse_yaml(files[0], **parse_kwargs)
        else:
            yamlfile.parse_yaml(files, **parse_kwargs)


class YamlTestBaseError(YamlTestBase):
    r"""Test error for yamlfile."""
    
    _error = None

    @pytest.fixture(scope="class")
    def error(self):
        return self._error

    def test_parse_yaml(self, parse_kwargs, error, nfiles, files):
        r"""Test error reading & parsing yaml."""
        if (error is None) or (nfiles == 0):
            return
        with pytest.raises(error):
            yamlfile.parse_yaml(files, parse_kwargs)


class TestJSONModelOnly(YamlTestBase):
    r"""Test parsing of different numbers/styles of models from JSON."""
    _use_json = True
    _contents = (['{"models":',
                  '   [{"name": "modelA",',
                  '     "driver": "GCCModelDriver",',
                  '     "args": "./src/modelA.c"}]}'],)
        

class TestYamlModelOnly(YamlTestBase):
    r"""Test parsing of different numbers/styles of models."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c'],
                 ['model:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c'],
                 ['model:',
                  '  name: modelC',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelC.c'],
                 ['models:',
                  '  name: modelD',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelD.c',
                  '  env:',
                  '    TEST_VAR: 1'], )


class TestYamlServerClient(YamlTestBase):
    r"""Test specification of server/client models."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    is_server: True'],
                 ['model:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    client_of: modelA'],)


class TestYamlIODrivers(YamlTestBase):
    r"""Test full specification of IO drivers."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - name: inputA',
                  '        driver: FileInputDriver',
                  '        translator: %s:direct_translate' % __name__,
                  '        onexit: printStatus',
                  '        args: {{ %s }}' % _yaml_env,
                  '    outputs:',
                  '      - name: outputA',
                  '        driver: FileOutputDriver',
                  '        translator: %s:direct_translate' % __name__,
                  '        onexit: printStatus',
                  '        args: fileA.txt',
                  '      - name: outputA2',
                  '        driver: OutputDriver',
                  '        translator: %s:direct_translate' % __name__,
                  '        onexit: printStatus',
                  '        args: A_to_B'],
                 ['model:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    input:',
                  '      - name: inputB',
                  '        driver: FileInputDriver',
                  '        args: {{ %s }}' % _yaml_env,
                  '      - name: inputB2',
                  '        driver: InputDriver',
                  '        args: A_to_B'],
                 ['model:',
                  '  name: modelC',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelC.c',
                  '  input:',
                  '    name: inputC',
                  '    driver: FileInputDriver',
                  '    args: {{ %s }}' % _yaml_env],
                 ['models:',
                  '  name: modelD',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelD.c',
                  '  inputs:',
                  '    name: inputD',
                  '    driver: FileInputDriver',
                  '    args: {{ %s }}' % _yaml_env], )


class TestYamlConnection(YamlTestBase):
    r"""Test connection between I/O channels."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlInclude(YamlTestBase):
    r"""Test connection between I/O channels in yaml included from one."""
    _contents = (['include: {{ %s0 }}' % _yaml_env,
                  'models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA'],)
    _include = (['models:',
                 '  - name: modelB',
                 '    driver: GCCModelDriver',
                 '    args: ./src/modelB.c',
                 '    outputs:',
                 '      - outputB'],)


class TestYamlIODatatype(YamlTestBase):
    r"""Test specification of datatype via schema."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - name: inputA',
                  '        type: object',
                  '        properties:',
                  '          a: int',
                  "          b: {'type': float, 'units': 'cm'}",
                  '',
                  'connections:',
                  "  - from: ['outputB::0', 'outputB::1']",
                  "    to: ['inputA::a', 'inputA::b']"],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - name: outputB',
                  '        type: [int, float]'],)


class TestYamlConnectionFork(YamlTestBase):
    r"""Test connection between I/O channels."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - inputs: ',
                  '      - outputB',
                  '      - outputC',
                  '    output: inputA',
                  '  - input: outputA',
                  '    outputs:',
                  '      - inputB',
                  '      - inputC'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    inputs:',
                  '      - inputB',
                  '    outputs:',
                  '      - outputB'],
                 ['models:',
                  '  - name: modelC',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelC.c',
                  '    inputs:',
                  '      - inputC',
                  '    outputs:',
                  '      - outputC'],)


class TestYamlConnectionTranslator(YamlTestBase):
    r"""Test connection between I/O channels."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA',
                  '    translator: %s:direct_translate' % __name__],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlConnectionInputFile(YamlTestBase):
    r"""Test connection with File."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input:',
                  '      - {{ %s }}' % _yaml_env,
                  '    read_meth: all',
                  '    output: inputA',
                  '  - input: outputA',
                  '    output:',
                  '      - output.txt',
                  '    write_meth: all'],)


class TestYamlConnectionInputFile_wait(YamlTestBase):
    r"""Test connection with File where wait_for_creation specified."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input:',
                  '      - {{ %s }}' % _yaml_env,
                  '    read_meth: all',
                  '    output: inputA',
                  '  - input: outputA',
                  '    output:',
                  '      - output.txt',
                  '    write_meth: all'],)


class TestYamlConnectionInputAsciiFile(YamlTestBase):
    r"""Test connection with AsciiFile."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: invalid.txt',
                  '    output: inputA',
                  '    wait_for_creation: 4',
                  '    read_meth: line',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: line'],)


class TestYamlConnectionInputAsciiTable(YamlTestBase):
    r"""Test connection with AsciiTable."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: table',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: table'],)


class TestYamlConnectionInputAsciiTableArray(YamlTestBase):
    r"""Test connection with AsciiTable as array."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: table_array',
                  '    field_units: name,count,size',
                  '    format_str: "%5s\t%d\t%f\n"',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: table_array',
                  '    field_units: name,count,size',
                  '    format_str: "%5s\t%d\t%f\n"'], )


class TestYamlConnectionInputPickle(YamlTestBase):
    r"""Test connection with Pickle."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: pickle',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: pickle'], )


class TestYamlConnectionInputPandas(YamlTestBase):
    r"""Test connection with Pandas csv."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: pandas',
                  '    field_units: name,count,size',
                  '    format_str: "%5s\t%d\t%f\n"',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: pandas',
                  '    field_units: name,count,size',
                  '    format_str: "%5s\t%d\t%f\n"'], )


class TestYamlConnectionInputAsciiMap(YamlTestBase):
    r"""Test connection with AsciiMap."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: map',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: map'], )


class TestYamlConnectionInputPly(YamlTestBase):
    r"""Test connection with Ply file."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: ply',
                  '  - input: outputA',
                  '    output: output.ply',
                  '    write_meth: ply'], )


class TestYamlConnectionInputObj(YamlTestBase):
    r"""Test connection with Obj file."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: obj',
                  '  - input: outputA',
                  '    output: output.obj',
                  '    write_meth: obj'], )


class TestYamlModelFunction(YamlTestBase):
    r"""Test when missing input/output connections should be routed to a function."""
    _parse_kwargs = {'complete_partial': True}
    _contents = (['models:',
                  '  - name: modelA',
                  '    language: c',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  '  - name: modelB',
                  '    language: c',
                  '    args: ./src/modelB.c',
                  '    inputs:',
                  '      - inputB',
                  '    outputs:',
                  '      - outputB',
                  '',
                  'connections:',
                  '  - input: outputA',
                  '    output: inputB'], )


class TestYamlConnectionDefaultValue(YamlTestBase):
    r"""Test use of default_value parameter."""
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - name: inputA',
                  '        default_value: test'],)


class TestYamlModelOnlyKwarg(YamlTestBase):
    r"""Test parsing with model_only=True."""
    _parse_kwargs = {'model_only': True}
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - name: inputA'],)


class TestYamlModelSubmission(YamlTestBase):
    r"""Test parsing with model_submission=True."""
    _parse_kwargs = {'model_submission': True}
    _contents = (['model:',
                  '  - name: FakeModel',
                  '    repository_url: https://github.com/cropsinsilico/'
                  'example-fakemodel',
                  '    repository_commit: e4bc7932c3c0c68fb3852cfb864777ca64cba448',
                  '    description: Example model submission',
                  '    language: python',
                  '    args:',
                  '      - ./src/fakemodel.py',
                  '      - --yggdrasil',
                  '    inputs:',
                  '      - name: photosynthesis_rate',
                  '        default_file:',
                  '          name: ./Input/input.txt',
                  '          filetype: table',
                  '        datatype:',
                  '          type: bytes',
                  '    outputs:',
                  '      - name: growth_rate',
                  '        default_file:',
                  '          name: ./Output/output.txt',
                  '          filetype: table',
                  '        datatype:',
                  '          type: bytes',
                  ],)

    @pytest.fixture(autouse=True)
    def cleanup_git_repo(self):
        r"""Remove the git repository."""
        yield
        import git
        git.rmtree("cropsinsilico")


class TestYamlServerNoClient(YamlTestBaseError):
    r"""Test error when is_server is set but there arn't any clients."""
    _error = yamlfile.YAMLSpecificationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    is_server: True'],)


class TestYamlClientNoServer(YamlTestBaseError):
    r"""Test error when client_of is set but there arn't any servers by that name."""
    _error = yamlfile.YAMLSpecificationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    client_of: modelB'],)


class TestYamlServerFunction(YamlTestBaseError):
    r"""Test error raised when both is_server and function are set, but
    there is more than one input/output."""
    _error = yamlfile.YAMLSpecificationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    function: fake',
                  '    is_server: True',
                  '    inputs: [inputA, inputB]'],
                 ['model:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    client_of: modelA'],)


class TestYamlServerDictNoInput(YamlTestBaseError):
    r"""Test error raised when is_server is a dictionary but the referenced
    input channel cannot be located."""
    _error = yamlfile.YAMLSpecificationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    function: fake',
                  '    is_server:',
                  '      input: A',
                  '      output: B',
                  '    outputs: B'],
                 ['model:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    client_of: modelA'],)


class TestYamlServerDictNoOutput(YamlTestBaseError):
    r"""Test error raised when is_server is a dictionary but the referenced
    output channel cannot be located."""
    _error = yamlfile.YAMLSpecificationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    function: fake',
                  '    is_server:',
                  '      input: A',
                  '      output: B',
                  '    inputs: A'],
                 ['model:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    client_of: modelA'],)


class TestYamlComponentError(YamlTestBaseError):
    r"""Test error for non-dictionary component."""
    _error = ValidationError
    _contents = (['models: error'],)


class TestYamlDuplicateKeyError(YamlTestBaseError):
    r"""Test error when there are duplicates of the same key in a map."""
    _error = ConstructorError
    _contents = (['model:',
                  '  name: modelA',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelA.c',
                  '  inputs:',
                  '    - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA',
                  '',
                  'model:',
                  '  name: modelB',
                  '  driver: GCCModelDriver',
                  '  args: ./src/modelB.c',
                  '  outputs:',
                  '    - outputB'],)


class TestYamlDuplicateError(YamlTestBaseError):
    r"""Test error when there are two components with the same name."""
    _error = yamlfile.YAMLSpecificationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c'],)


class TestYamlConnectionError(YamlTestBaseError):
    r"""Test error when there is not connection for a model I/O channel."""
    _error = yamlfile.YAMLSpecificationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA'],)


class TestYamlConnectionError_forkin(YamlTestBaseError):
    r"""Test error when there is not connection for a fork input channel."""
    _error = yamlfile.YAMLSpecificationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  'connections:',
                  '  - input:',
                  '      - output1',
                  '      - output2',
                  '    output: inputA'],)


# Error not raised as both outputs could be files
# class TestYamlConnectionError_forkout(YamlTestBaseError):
#     r"""Test error when there is not connection for a fork output channel."""
#     _error = RuntimeError
#     _contents = (['models:',
#                   '  - name: modelA',
#                   '    driver: GCCModelDriver',
#                   '    args: ./src/modelA.c',
#                   '    outputs:',
#                   '      - outputA',
#                   'connections:',
#                   '  - input: outputA',
#                   '    output:',
#                   '      - input1',
#                   '      - input2'],)


class TestYamlConnectionError_readmeth(YamlTestBaseError):
    r"""Test error when read_meth is specified for non-file."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA',
                  '    read_meth: pickle'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlConnectionError_writemeth(YamlTestBaseError):
    r"""Test error when write_meth is specified for non-file."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA',
                  '    write_meth: pickle'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlMissingModelArgsError(YamlTestBaseError):
    r"""Test error when there is a missing arguments to a model."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    inputs:',
                  '      name: inputA',
                  '      driver: FileInputDriver',
                  '      args: {{ %s }}' % _yaml_env],)


class TestYamlMissingIOArgsError_input(YamlTestBaseError):
    r"""Test error when there is a missing arguments to an input driver."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      name: inputA',
                  '      driver: FileInputDriver'],)


class TestYamlMissingIOArgsError_output(YamlTestBaseError):
    r"""Test error when there is a missing arguments to an output driver."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    outputs:',
                  '      name: outputA',
                  '      driver: FileOutputDriver'],)


class TestYamlMissingConnArgsError(YamlTestBaseError):
    r"""Test error when there is a missing arguments to a connection."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c',
                  '    outputs:',
                  '      - outputB'],)


class TestYamlMissingConnInputError(YamlTestBaseError):
    r"""Test error when there is no model output matching connection input."""
    _error = RuntimeError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c'],)


class TestYamlMissingConnInputFileError(YamlTestBaseError):
    r"""Test error when there is no file for missing connection input."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '',
                  'connections:',
                  '  - input: outputB',
                  '    output: inputA'],
                 ['models:',
                  '  - name: modelB',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelB.c'],)


class TestYamlMissingConnIOError(YamlTestBaseError):
    r"""Test error when there is no model input/output matching connection."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA'],)


class TestYamlConnectionInputFileReadMethError(YamlTestBaseError):
    r"""Test error for invalid read_meth."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '    read_meth: invalid',
                  '  - input: outputA',
                  '    output: output.txt'],)


class TestYamlConnectionInputFileWriteMethError(YamlTestBaseError):
    r"""Test error for invalid write_meth."""
    _error = ValidationError
    _contents = (['models:',
                  '  - name: modelA',
                  '    driver: GCCModelDriver',
                  '    args: ./src/modelA.c',
                  '    inputs:',
                  '      - inputA',
                  '    outputs:',
                  '      - outputA',
                  '',
                  'connections:',
                  '  - input: {{ %s }}' % _yaml_env,
                  '    output: inputA',
                  '  - input: outputA',
                  '    output: output.txt',
                  '    write_meth: invalid'],)



import pytest
import os
import tempfile
from yggdrasil import constants
from tests.drivers.test_ConnectionDriver import TestInputDriver as base_class


_filetypes = sorted(
    [x for x in constants.COMPONENT_REGISTRY['file']['subtypes'].keys()
     if x not in ['pandas']])


class TestFileInputDriver(base_class):
    r"""Test class for FileInputDriver."""

    parametrize_filetype = _filetypes

    test_send_recv = None
    test_send_recv_nolimit = None
    
    @pytest.fixture(scope="class")
    def component_subtype(self):
        r"""Subtype of component being tested."""
        return 'file_input'

    @pytest.fixture(scope="class")
    def filetype(self, request):
        r"""str: Name of the file type being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def commtype(self, filetype):
        r"""str: Name of the communicator being tested."""
        return filetype

    @pytest.fixture
    def instance_args(self, name, filepath):
        r"""Arguments for a new instance of the tested class."""
        return (name, filepath)

    @pytest.fixture
    def send_comm_kwargs(self, instance, icomm_name):
        r"""dict: Keyword arguments for send comm."""
        out = instance.icomm.opp_comm_kwargs()
        out['append'] = True
        return out

    @pytest.fixture
    def filepath(self, name, icomm_python_class, contents_to_write):
        r"""str: Path to the test file."""
        out = os.path.join(
            tempfile.gettempdir(),
            f'{name}_input{icomm_python_class._default_extension}')
        with open(out, 'wb') as fd:
            fd.write(contents_to_write)
        try:
            yield out
        finally:
            if os.path.isfile(out):
                os.remove(out)
    
    @pytest.fixture(scope="class")
    def contents_to_write(self, testing_options):
        r"""str: Contents that should be written to the file."""
        return testing_options['contents']

    @pytest.fixture
    def assert_before_stop(self, instance, recv_message_list,
                           recv_comm, testing_options, unyts_equality_patch,
                           pandas_equality_patch):
        r"""Assertions to make before stopping the driver instance."""
        def assert_before_stop_w():
            instance.sleep()
            recv_message_list(recv_comm, testing_options['recv'])
        return assert_before_stop_w
    
    # These are disabled to prevent writting extraneous data
    @pytest.fixture(scope="class")
    def run_before_terminate(self):
        r"""Commands to run while the instance is running, before terminate."""
        def run_before_terminate_w():
            pass
        return run_before_terminate_w


class TestPandasInputDriver(TestFileInputDriver):
    r"""Test class for FileInputDriver reading from a pandas table."""
    
    @pytest.fixture(scope="class")
    def filetype(self):
        r"""str: Name of the file type being tested."""
        return 'pandas'

    @pytest.fixture(scope="class")
    def options(self):
        r"""Arguments that should be provided when getting testing options."""
        return {'not_as_frames': True}


class TestAsciiTableArrayInputDriver(TestFileInputDriver):
    r"""Test class for FileInputDriver reading from a table as an array."""

    @pytest.fixture(scope="class")
    def filetype(self):
        r"""str: Name of the file type being tested."""
        return 'table'

    @pytest.fixture(scope="class")
    def options(self):
        r"""Arguments that should be provided when getting testing options."""
        return {'array_columns': True}



import pytest
import os
import numpy as np
from yggdrasil import units, tools
from yggdrasil.communication import AsciiTableComm
from yggdrasil.metaschema.properties.ScalarMetaschemaProperties import (
    data2dtype)
from tests.communication.test_FileComm import TestFileComm as base_class


def test_AsciiTableComm_nofmt():
    r"""Test read of asciitable without format."""
    test_file = os.path.join(os.getcwd(), 'temp_file.txt')
    rows = [('one', 1, 1.0), ('two', 2, 2.0), ('three', 3, 3.0)]
    lines = [('%5s\t%d\t%f\n' % r) for r in rows]
    contents = (''.join(lines)).encode("utf-8")
    with open(test_file, 'wb') as fd:
        fd.write(contents)
    inst = AsciiTableComm.AsciiTableComm('test', test_file, direction='recv')
    inst.open()
    for ans in rows:
        flag, x = inst.recv_dict()
        assert(flag)
        irow = [e for e in ans]
        irow[0] = irow[0].encode("utf-8")
        idict = {'f%d' % i: irow[i] for i in range(len(irow))}
        # irow = tuple(irow)
        assert(x == idict)
    flag, x = inst.recv()
    assert(not flag)
    inst.close()
    os.remove(test_file)


class TestAsciiTableComm(base_class):
    r"""Test for AsciiTableComm communication class."""

    test_send_recv_comment = None

    @pytest.fixture(scope="class", autouse=True)
    def filetype(self):
        r"""Communicator type being tested."""
        return "table"

    @pytest.fixture(scope="class")
    def map_sent2recv(self, testing_options):
        r"""Factory for method to convert sent messages to received."""
        def wrapped_map_sent2recv(obj):
            if (not isinstance(obj, bytes)) or (obj != tools.YGG_MSG_EOF):
                field_units = testing_options.get('field_units', None)
                if field_units:
                    if isinstance(obj, dict):
                        return {k: units.add_units(v, u, dtype=data2dtype(v))
                                for (k, v), u in zip(obj.items(), field_units)}
                    elif isinstance(obj, (list, tuple)):
                        return [units.add_units(x, u, dtype=data2dtype(x))
                                for x, u in zip(obj, field_units)]
            return obj
        return wrapped_map_sent2recv


@pytest.mark.usefixtures("unyts_equality_patch")
class TestAsciiTableComm_AsArray(TestAsciiTableComm):
    r"""Test for AsciiTableComm communication class."""

    @pytest.fixture(scope="class", autouse=True)
    def options(self):
        r"""Arguments that should be provided when getting testing options."""
        return {'array_columns': True}


class TestAsciiTableComm_single(TestAsciiTableComm):
    r"""Test for AsciiTableComm communication class with field names sent."""

    @pytest.fixture(scope="class")
    def testing_options(self):
        r"""Testing options."""
        nele = 5
        dtype = np.dtype(dict(formats=['float'], names=['f0']))
        arr1 = np.zeros((nele, ), dtype)
        arr2 = np.ones((nele, ), dtype)
        out = {'kwargs': {'as_array': True, 'field_names': ['f0']},
               'contents': (
                   b'# f0\n# %g\n'
                   + nele * b'0\n' + nele * b'1\n'),
               'send': [[arr1['f0']], [arr2['f0']]],
               'recv': [[np.hstack([arr1, arr2])['f0']]],
               'recv_partial': [[[arr1['f0']]], [[arr2['f0']]]],
               'dict': {'f0': arr1['f0']},
               'objects': [[arr1['f0']], [arr2['f0']]]}
        out['msg'] = out['send'][0]
        out['msg_array'] = arr1
        return out

    def test_send_dict_default(self, send_comm, recv_comm, do_send_recv,
                               testing_options):
        r"""Test automated conversion of dictionary to pandas data frame."""
        do_send_recv(send_comm, recv_comm,
                     send_params={'message': testing_options['dict']},
                     recv_params={'message': testing_options['msg']})



import pprint
import numpy as np
from yggdrasil.metaschema.datatypes import get_type_class


def get_test_data(typename):
    r"""Determine a test data set for the specified type.

    Args:
        typename (str): Name of datatype.

    Returns:
        object: Example of specified datatype.

    """
    typeclass = get_type_class(typename)
    return typeclass.get_test_data()


def check_received_data(typename, x_recv):
    r"""Check that the received message is equivalent to the
    test data for the specified type.

    Args:
        typename (str): Name of datatype.
        x_recv (object): Received object.

    Raises:
        AssertionError: If the received message is not equivalent
            to the received message.

    """
    x_sent = get_test_data(typename)
    print('RECEIVED:')
    pprint.pprint(x_recv)
    print('EXPECTED:')
    pprint.pprint(x_sent)
    if isinstance(x_sent, np.ndarray):
        np.testing.assert_array_equal(x_recv, x_sent)
    else:
        assert(x_recv == x_sent)



import os


def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        input_files='input.txt',
        expected_output_files=os.path.join('Output', 'expected.txt'),
        output_dir=None,
        output_files=['output.txt'])



import pytest
import numpy as np
import pandas
from tests.communication.test_AsciiTableComm import (
    TestAsciiTableComm as base_class)


@pytest.mark.usefixtures("pandas_equality_patch")
class TestPandasFileComm(base_class):
    r"""Test for PandasFileComm communication class."""

    @pytest.fixture(scope="class", autouse=True)
    def filetype(self):
        r"""Communicator type being tested."""
        return "pandas"

    @pytest.fixture(scope="class", autouse=True,
                    params=[{}, {'no_names': True}])
    def options(self, request):
        r"""Arguments that should be provided when getting testing options."""
        return request.param

    @pytest.fixture(scope="class")
    def map_sent2recv(self, testing_options):
        r"""Factory for method to convert sent messages to received."""
        def wrapped_map_sent2recv(obj):
            return obj
        return wrapped_map_sent2recv


class TestPandasFileComm_single(TestPandasFileComm):
    r"""Test for PandasFileComm communication class with field names sent."""

    @pytest.fixture(scope="class", autouse=True)
    def options(self):
        r"""Arguments that should be provided when getting testing options."""
        return {}
    
    @pytest.fixture(scope="class")
    def testing_options(self):
        r"""Testing options."""
        nele = 5
        dtype = np.dtype(dict(formats=['float'], names=['f0']))
        arr1 = np.zeros((nele, ), dtype)
        arr2 = np.ones((nele, ), dtype)
        pd1 = pandas.DataFrame(arr1)
        pd2 = pandas.DataFrame(arr2)
        pdcat = pandas.DataFrame(np.hstack([arr1, arr2]))
        out = {'kwargs': {},
               'contents': (b'f0\n' + nele * b'0.0\n' + nele * b'1.0\n'),
               'send': [pd1, pd2],
               'recv': [pdcat],
               'recv_partial': [[pd1], [pdcat]],
               'dict': {'f0': arr1['f0']},
               'objects': [pd1, pd2]}
        out['msg'] = out['send'][0]
        out['msg_array'] = arr1
        return out

    def test_send_dict_default(self, send_comm, recv_comm, testing_options,
                               do_send_recv):
        r"""Test automated conversion of dictionary to pandas data frame."""
        do_send_recv(send_comm, recv_comm,
                     send_params={'message': testing_options['dict']},
                     recv_params={'message': testing_options['msg']})



def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        output_dir='Output',
        output_files=['root_output.txt',
                      'shoot_output.txt'],
        skip_check_results=True)



import pytest
import flaky
from tests.communication.test_CommBase import TestComm as base_class
from yggdrasil.communication import new_comm
from yggdrasil.communication.RMQComm import RMQComm
from tests import timeout_decorator


@flaky.flaky
@timeout_decorator(timeout=60)
class TestRMQAsyncComm(base_class):
    r"""Test for RMQAsyncComm communication class."""

    @pytest.fixture(scope="class", autouse=True)
    def commtype(self):
        r"""Communicator type being tested."""
        return "rmq_async"

    @pytest.fixture(autouse=True)
    def ensure_comm(self, send_comm, recv_comm):
        pass

    def test_reconnect_recv(self, send_comm, recv_comm, do_send_recv):
        r"""Test reconnect after unexpected disconnect of recv comm."""
        send_comm.printStatus()
        recv_comm.printStatus()
        do_send_recv(send_comm, recv_comm)
        recv_comm.connection.close(reply_code=100,
                                   reply_text="Test shutdown")
        recv_comm._reconnecting.started.wait(5)
        recv_comm._reconnecting.stopped.wait(5)
        assert(recv_comm.times_connected > 1)
        assert(recv_comm._reconnecting.has_stopped())
        recv_comm._opening.stopped.wait(5)
        do_send_recv(send_comm, recv_comm)
        send_comm.printStatus()
        recv_comm.printStatus()


class TestRMQAsyncCommNamedQueue(TestRMQAsyncComm):
    r"""Test for RMQAsyncComm communication class with a named queue."""

    test_attributes = None
    test_cleanup_comms = None
    test_drain_messages = None
    test_empty_obj_recv = None
    test_eof = None
    test_eof_no_close = None
    test_error_name = None
    test_error_recv = None
    test_error_send = None
    test_invalid_direction = None
    test_purge = None
    test_recv_nomsg = None
    test_send_recv = None
    test_send_recv_after_close = None
    test_send_recv_array = None
    test_send_recv_dict = None
    test_send_recv_dict_names = None
    test_send_recv_filter_eof = None
    test_send_recv_filter_pass = None
    test_send_recv_filter_recv_filter = None
    test_send_recv_filter_send_filter = None
    test_send_recv_nolimit = None
    test_send_recv_raw = None
    test_work_comm = None

    @pytest.fixture(scope="class")
    def testing_options(self, python_class, options):
        r"""Testing options."""
        out = python_class.get_testing_options(**options)
        out['kwargs'].update(queue='test_queue')
        return out
    
    def test_reconnect_send(self, send_comm, recv_comm, do_send_recv):
        r"""Test reconnect after unexpected disconnect of send comm."""
        send_comm.start_run_thread()
        send_comm.printStatus()
        recv_comm.printStatus()
        do_send_recv(send_comm, recv_comm)
        send_comm.connection.close(reply_code=100,
                                   reply_text="Test shutdown")
        send_comm._reconnecting.started.wait(5)
        send_comm._reconnecting.stopped.wait(10)
        assert(send_comm.times_connected > 1)
        assert(send_comm._reconnecting.has_stopped())
        do_send_recv(send_comm, recv_comm)
        send_comm.printStatus()
        recv_comm.printStatus()


@pytest.mark.skipif(RMQComm.is_installed(language='python'),
                    reason="RMQ Server running")
def test_not_running():
    r"""Test raise of an error if a RMQ server is not running."""
    comm_kwargs = dict(commtype='rmq_async', direction='send',
                       reverse_names=True)
    with pytest.raises(RuntimeError):
        new_comm('test', **comm_kwargs)



import pytest
from yggdrasil import languages, constants, platform


def test_get_language_dir():
    r"""Test the get_language_dir method."""
    with pytest.raises(ValueError):
        languages.get_language_dir('invalid')
    test_lang = constants.LANGUAGES_WITH_ALIASES['all']
    test_skip = ['make', 'cmake', 'executable', 'timesync', 'osr', 'mpi']
    for lang in test_lang:
        if lang in test_skip:
            continue
        languages.get_language_dir(lang)


def test_get_language_ext():
    r"""Test the get_language_ext method."""
    test_lang = constants.LANGUAGES_WITH_ALIASES['all']
    for lang in test_lang:
        if ((((lang == 'executable') and (not platform._is_win))
             or (lang in ['dummy', 'mpi']))):
            with pytest.raises(ValueError):
                languages.get_language_ext(lang)
            languages.get_language_ext(lang, default='')
        else:
            languages.get_language_ext(lang)



import pytest
import os
import jsonschema
from yggdrasil import schema
from yggdrasil.communication import new_comm
from tests.communication.test_CommBase import TestComm as base_class


def test_wait_for_creation():
    r"""Test FileComm waiting for creation."""
    msg_send = b'Test message\n'
    name = 'temp_file_create.txt'
    kwargs = {'in_temp': True, 'commtype': 'binary', 'dont_open': True}
    # kwargs = {'wait_for_creation': 5, 'in_temp': True, commtype='binary'}
    send_instance = new_comm(name, direction='send', **kwargs)
    recv_instance = new_comm(name, direction='recv',
                             wait_for_creation=5.0, **kwargs)
    if os.path.isfile(send_instance.address):
        os.remove(send_instance.address)
    
    def open_and_send(inst, msg):
        inst.open()
        flag = inst.send(msg)
        return flag
    
    send_instance.sched_task(0.5, open_and_send, args=[send_instance, msg_send],
                             store_output=True)
    recv_instance.open()
    T = recv_instance.start_timeout(recv_instance.wait_for_creation)
    while (not T.is_out) and (send_instance.sched_out is None):  # pragma: debug
        recv_instance.sleep()
    recv_instance.stop_timeout()
    assert(send_instance.sched_out)
    flag, msg_recv = recv_instance.recv()
    assert(flag)
    assert(msg_recv == msg_send)
    send_instance.close()
    recv_instance.close()
    recv_instance.remove_file()


_filetypes = sorted([x for x in schema.get_schema()['file'].subtypes
                     if x not in ['ascii', 'table', 'pandas']])


@pytest.mark.usefixtures("unyts_equality_patch")
@pytest.mark.suite("files", ignore="comms")
class TestFileComm(base_class):
    r"""Test for FileComm communication class."""

    _component_type = 'file'
    parametrize_filetype = _filetypes

    test_send_recv_nolimit = None
    test_work_comm = None
    test_send_recv_raw = None
    
    @pytest.fixture(scope="class", autouse=True)
    def component_subtype(self, filetype):
        r"""Subtype of component being tested."""
        return filetype

    @pytest.fixture(scope="class", autouse=True)
    def filetype(self, request):
        r"""Communicator type being tested."""
        return request.param

    @pytest.fixture(scope="class", autouse=True)
    def commtype(self, filetype):
        r"""Communicator type being tested."""
        return filetype

    @pytest.fixture(scope="class", autouse=True)
    def use_async(self):
        r"""Whether communicator should be asynchronous or not."""
        return False
    
    def get_send_comm_kwargs(self, *args, **kwargs):
        r"""Get keyword arguments for creating a send comm."""
        kwargs['in_temp'] = True
        return super(TestFileComm, self).get_send_comm_kwargs(
            *args, **kwargs)

    @pytest.fixture(scope="class")
    def close_comm(self, close_comm):
        r"""Remove a comm."""
        def close_comm_w(comm, dont_remove_file=False):
            close_comm(comm)
            if (comm.direction == 'send') and (not dont_remove_file):
                comm.remove_file()
        return close_comm_w

    @pytest.fixture
    def global_send_comm(self, send_comm):
        r"""Communicator for sending messages."""
        return send_comm

    @pytest.fixture
    def global_recv_comm(self, recv_comm):
        r"""Communicator for receiving messages."""
        return recv_comm

    @pytest.fixture
    def global_comm(self, global_recv_comm):
        r"""Global communicator."""
        return global_recv_comm
    
    @pytest.fixture(scope="class")
    def maxMsgSize(self):
        r"""int: Maximum message size."""
        return 0

    def test_file_size(self, global_recv_comm):
        r"""Test file_size method."""
        global_recv_comm.file_size

    def test_send_recv_eof_no_close(self, send_comm, recv_comm, do_send_recv,
                                    timeout):
        r"""Test send/recv of EOF message with no close."""
        recv_comm.close_on_eof_recv = False
        do_send_recv(send_comm, recv_comm,
                     send_params={'method': 'send_eof'},
                     recv_params={'flag': True,
                                  'skip_wait': True,
                                  'kwargs': {'timeout': timeout}})

    def test_send_recv_filter_eof(self, run_once, filtered_comms, send_comm,
                                  recv_comm, do_send_recv, timeout):
        r"""Test send/recv of EOF with filter."""
        do_send_recv(send_comm, recv_comm,
                     send_params={'method': 'send_eof'},
                     recv_params={'flag': False,
                                  'skip_wait': True,
                                  'kwargs': {'timeout': timeout}})
        assert(recv_comm.is_closed)

    def test_send_recv_filter_send_filter(self, filtered_comms,
                                          msg_filter_send, send_comm,
                                          recv_comm, polling_interval,
                                          do_send_recv):
        r"""Test send/recv with filter that blocks send."""
        do_send_recv(send_comm, recv_comm, msg_filter_send,
                     recv_params={'message': recv_comm.eof_msg,
                                  'flag': False,
                                  'skip_wait': True,
                                  'kwargs': {'timeout': polling_interval}})
        
    def test_send_recv_filter_recv_filter(self, filtered_comms,
                                          msg_filter_recv, send_comm,
                                          recv_comm, polling_interval,
                                          do_send_recv):
        r"""Test send/recv with filter that blocks recv."""
        # Wait if not async?
        do_send_recv(send_comm, recv_comm, msg_filter_recv,
                     recv_params={'message': recv_comm.eof_msg,
                                  'flag': False,
                                  'skip_wait': True,
                                  'kwargs': {'timeout': 10 * polling_interval}})
        
    def test_invalid_read_meth(self, name, commtype, use_async,
                               testing_options):
        r"""Test raise of error on invalid read_meth."""
        if commtype != 'binary':
            pytest.skip("Only run for commtype 'binary'")
        kws = self.get_send_comm_kwargs(
            commtype, use_async, testing_options, read_meth='invalid',
            skip_component_schema_normalization=False)
        with pytest.raises(jsonschema.ValidationError):
            new_comm(name, **kws)

    def test_append(self, uuid, commtype, use_async, testing_options,
                    send_comm, recv_comm, recv_message_list, close_comm):
        r"""Test open of file comm with append."""
        send_objects = testing_options['send']
        recv_objects = testing_options['recv']
        recv_objects_partial = testing_options['recv_partial']
        # Write to file
        flag = send_comm.send(send_objects[0])
        assert(flag)
        # Create temp file comms in append mode
        recv_kwargs = self.get_recv_comm_kwargs(
            commtype, send_comm, testing_options, append=True)
        new_inst_recv = new_comm('partial%s' % uuid, **recv_kwargs)
        send_kwargs = self.get_send_comm_kwargs(
            commtype, use_async, testing_options, append=True)
        new_inst_send = new_comm('append%s' % uuid, **send_kwargs)
        try:
            recv_message_list(new_inst_recv, recv_objects_partial[0],
                              break_on_empty=True)
            for i in range(1, len(send_objects)):
                flag = new_inst_send.send(send_objects[i])
                assert(flag)
                recv_message_list(new_inst_recv, recv_objects_partial[i],
                                  break_on_empty=True)
            # Read entire contents
            recv_message_list(recv_comm, recv_objects)
        finally:
            close_comm(new_inst_send, dont_remove_file=True)
            close_comm(new_inst_recv, dont_remove_file=True)
        # Check file contents
        if testing_options.get('exact_contents', True):
            with open(send_comm.address, 'rb') as fd:
                contents = fd.read()
            assert(contents == testing_options['contents'])

    def test_series(self, send_comm, recv_comm, do_send_recv):
        r"""Test sending/receiving to/from a series of files."""
        # Set up series
        fname = '%d'.join(os.path.splitext(send_comm.address))
        send_comm.close()
        recv_comm.close()
        send_comm.is_series = True
        recv_comm.is_series = True
        send_comm.address = fname
        recv_comm.address = fname
        send_comm.open()
        recv_comm.open()
        # Send/receive multiple messages
        nmsg = 2
        for i in range(nmsg):
            do_send_recv(send_comm, recv_comm)
        
    def test_remaining_bytes(self, send_comm, recv_comm):
        r"""Test remaining_bytes."""
        assert(send_comm.remaining_bytes == 0)
        recv_comm.close()
        assert(recv_comm.is_closed)
        assert(recv_comm.remaining_bytes == 0)

    def test_recv_nomsg(self, recv_comm, polling_interval):
        r"""Test recieve when there is no waiting message."""
        flag, msg_recv = recv_comm.recv(timeout=polling_interval)
        assert(not flag)
        assert(msg_recv == recv_comm.eof_msg)


class TestFileComm_readline(TestFileComm):
    r"""Test for FileComm communication class with read_meth = 'readline'."""

    @pytest.fixture(scope="class", autouse=True)
    def filetype(self):
        r"""Communicator type being tested."""
        return "ascii"

    @pytest.fixture(scope="class", autouse=True)
    def options(self, request):
        r"""Arguments that should be provided when getting testing options."""
        return {'read_meth': 'readline'}
    
    def get_recv_comm_kwargs(self, *args, **kwargs):
        r"""Get keyword arguments for creating a recv comm."""
        kwargs['read_meth'] = 'readline'
        return super(TestFileComm_readline, self).get_recv_comm_kwargs(
            *args, **kwargs)



import pytest
import os
from tests.examples import TestExample as base_class


@pytest.mark.extra_example
class TestExampleRpcFib(base_class):
    r"""Test the rpcFib example."""

    @pytest.fixture(scope="class")
    def example_name(self):
        r"""str: Name of example being tested."""
        return "rpcFib"
    
    @pytest.fixture(scope="class")
    def env(self):
        r"""dict: Environment variables set for the test."""
        return {'FIB_ITERATIONS': '3',
                'FIB_SERVER_SLEEP_SECONDS': '0.01'}
    
    @pytest.fixture
    def results(self, env):
        r"""Result that should be found in output files."""
        prev1 = 0
        prev2 = 1
        res = ''
        for i in range(int(env['FIB_ITERATIONS'])):
            next = prev1 + prev2
            res += 'fib(%2d<-) = %-2d<-\n' % ((i + 1), next)
            prev2 = prev1
            prev1 = next
        return [res]

    @pytest.fixture
    def output_files(self, tempdir):
        r"""Output file."""
        return [os.path.join(tempdir, 'fibCli.txt')]



import pytest
import numpy as np
from yggdrasil.serialize import AsciiTableSerialize
from yggdrasil import platform
from tests.serialize.test_DefaultSerialize import (
    TestDefaultSerialize as base_class)


def test_serialize_nofmt():
    r"""Test error on serialization without a format."""
    inst = AsciiTableSerialize.AsciiTableSerialize()
    inst.initialized = True
    test_msg = np.zeros((5, 5))
    with pytest.raises(RuntimeError):
        inst.serialize(test_msg)

    
def test_deserialize_nofmt():
    r"""Test error on deserialization without a format."""
    inst = AsciiTableSerialize.AsciiTableSerialize()
    test_msg = b'lskdbjs;kfbj'
    test_msg = inst.encoded_datatype.serialize(test_msg, metadata={})
    with pytest.raises(RuntimeError):
        inst.deserialize(test_msg)


def test_discover_header_no_header(tmpdir):
    r"""Test discover_header with single line, two columns, no header and
    non-default delimiter."""
    fd0 = tmpdir.join("temp_table.txt")
    fd0.write(b"13 berlin", 'wb')
    fd = fd0.open('rb')
    inst = AsciiTableSerialize.AsciiTableSerialize(delimiter=b' ')
    inst.deserialize_file_header(fd)
    if platform._is_win:  # pragma: windows
        assert(inst.format_str == b'%d %6s\n')
    else:
        assert(inst.format_str == b'%ld %6s\n')
    assert(inst.field_names == ('f0', 'f1'))


def test_discover_header_one_element(tmpdir):
    r"""Test discover_header with single line, single column, no header."""
    fd0 = tmpdir.join("temp_table.txt")
    fd0.write(b"berlin", 'wb')
    fd = fd0.open('rb')
    inst = AsciiTableSerialize.AsciiTableSerialize(delimiter=b' ')
    inst.deserialize_file_header(fd)
    assert(inst.format_str == b'%6s\n')
    assert(inst.field_names == ('f0',))


_options = [
    {},
    {'explicit_testing_options': {
        'kwargs': {'format_str': '%d\n'},
        'empty': [],
        'objects': [[1]],
        'extra_kwargs': {},
        'typedef': {'type': 'array',
                    'items': [{'type': 'int', 'precision': 32}]},
        'dtype': None,
        'is_user_defined': False}},
    {'array_columns': True},
]


class TestAsciiTableSerialize(base_class):
    r"""Test class for AsciiTableSerialize class."""

    @pytest.fixture(scope="class", autouse=True)
    def serializer(self):
        r"""str: Serializer being tested."""
        return "table"

    @pytest.fixture(scope="class", autouse=True, params=_options)
    def options(self, request):
        r"""Arguments that should be provided when getting testing options."""
        return request.param

    def test_field_specs(self, instance, testing_options, nested_approx):
        r"""Test field specifiers."""
        if not instance.initialized:
            instance.serialize(testing_options['objects'][0],
                               no_metadata=True)
        super(TestAsciiTableSerialize, self).test_field_specs(
            instance, testing_options, nested_approx)
        # Specific to this class
        if 'format_str' in testing_options:
            assert(instance.format_str
                   == testing_options['format_str'].encode("utf-8"))
        field_names = testing_options.get('field_names', None)
        assert(instance.field_names == field_names)
        field_units = testing_options.get('field_units', None)
        assert(instance.field_units == field_units)


class TestAsciiTableSerialize_object(TestAsciiTableSerialize):
    r"""Test class for AsciiTableSerialize class with object."""

    @pytest.fixture(scope="class", autouse=True, params=[{}])
    def options(self, request):
        r"""Arguments that should be provided when getting testing options."""
        return request.param

    @pytest.fixture(scope="class")
    def testing_options(self, python_class, options):
        r"""Testing options."""
        out = python_class.get_testing_options(**options)
        out.update(
            kwargs={'field_units': out['kwargs']['field_units']},
            format_str='%s\t%d\t%g\n',
            field_names=['%s_%s' % (k, x) for k, x
                         in zip('abc', out['field_names'])])
        out['objects'] = [{k: ix for k, ix in zip(out['field_names'], x)}
                          for x in out['objects']]
        for x, k2 in zip(out['typedef']['items'], out['field_names']):
            out['contents'].replace(x['title'].encode("utf-8"),
                                    k2.encode("utf-8"))
            x['title'] = k2
        return out

    @pytest.fixture
    def map_sent2recv(self, nested_approx, instance):
        r"""Factory for method to convert sent messages to received."""
        def wrapped_map_sent2recv(obj):
            return instance.datatype.coerce_type(
                nested_approx(obj), typedef=instance.typedef)
        return wrapped_map_sent2recv



import os
import uuid
import requests
from yggdrasil.communication import CommBase, NoMessages


def add_comm_server_to_app(app):
    r"""Add methods for handling send/receive calls server-side.

    Args:
        app (flask.Flask): Flask app to add methods to.

    """
    from flask import request
    app.queue = {}

    @app.route('/<client_id>/<model>/<channel>',
               methods=['GET', 'PUT', 'POST'])
    def queue(client_id, model, channel):
        r"""Respond to GET with queued message and add message from PUT
        or POST to the queue."""
        if request.method in ['PUT', 'POST']:
            # Queue a message when it is received from a client.
            app.queue.setdefault((client_id, model, channel), [])
            msg = request.get_data()
            app.queue[(client_id, model, channel)].append(msg)
            return b''
        else:
            # Return a message from the queue when requested by a client.
            if app.queue.get((client_id, model, channel), []):
                msg = app.queue[(client_id, model, channel)].pop(0)
                return msg
            else:
                return b''

    @app.route('/<client_id>/<model>/<channel>/size', methods=['GET'])
    def queue_size(client_id, model, channel):
        r"""Return the size of the message queue."""
        return str(len(app.queue.get((client_id, model, channel), [])))

    @app.route('/<client_id>/<model>/<channel>/purge', methods=['GET'])
    def queue_purge(client_id, model, channel):
        r"""Return the size of the message queue."""
        app.queue[(client_id, model, channel)] = []
        return b''

    @app.route('/<client_id>/<model>/<channel>/remove', methods=['GET'])
    def queue_remove(client_id, model, channel):
        r"""Remove a queue."""
        app.queue.pop((client_id, model, channel), None)
        return b''


class RESTComm(CommBase.CommBase):
    r"""Class for handling I/O via a RESTful API. The provided address should
    be an HTTP address to a server running a flask app that has been equipped
    to respond to send/receive calls via the add_comm_server_to_app method.

    Args:
        params (dict, optional): Parameters that should be passed via URL.
            Defaults to None and is ignored.
        cookies (dict, optional): Cookies to send to the server. Defaults to
            None and is ignored.
        **kwargs: Additional keyword arguments will be passed to the base
            class.

    """

    _commtype = 'rest'
    _schema_subtype_description = 'RESTful API.'
    _schema_properties = {
        'client_id': {'type': 'string'},
        'params': {'type': 'object'},
        'cookies': {'type': 'object'},
        'host': {'type': 'string', 'default': 'http://localhost:{port}'},
        'port': {'type': 'int'}}
    _maxMsgSize = 2048  # Based on limit for GET requests on most servers

    def __init__(self, *args, **kwargs):
        self._is_open = False
        super(RESTComm, self).__init__(*args, **kwargs)

    def atexit(self):  # pragma: debug
        r"""Close operations."""
        if self.direction == 'send':
            self.linger()
        super(RESTComm, self).atexit()
        
    @property
    def is_open(self):
        r"""bool: True if the connection is open."""
        return self._is_open
    
    def open(self, *args, **kwargs):
        r"""Open the connection."""
        super(RESTComm, self).open(*args, **kwargs)
        self._is_open = True

    def bind(self, *args, **kwargs):
        r"""Bind to address based on information provided."""
        if self.address == 'address':
            client_id = self.client_id
            if client_id is None:
                client_id = str(uuid.uuid4()).split('-')[0]
            model = self.partner_model
            if model is None:
                model = str(uuid.uuid4()).split('-')[0]
            if self.port is None:
                self.port = int(os.environ.get("PORT", 5000))
            if not self.host.endswith('/'):
                self.host += '/'
            if '{port}' in self.host:
                self.host = self.host.format(port=self.port)
            host = self.host
            name = self.name_base
            assert(client_id is not None)
            self.address = f'{host}{client_id}/{model}/{name}'
        return super(RESTComm, self).bind(*args, **kwargs)

    def opp_comm_kwargs(self, for_yaml=False):
        r"""Get keyword arguments to initialize communication with opposite
        comm object.

        Args:
            for_yaml (bool, optional): If True, the returned dict will only
                contain values that can be specified in a YAML file. Defaults
                to False.

        Returns:
            dict: Keyword arguments for opposite comm object.

        """
        from yggdrasil.services import _service_host_env
        out = super(RESTComm, self).opp_comm_kwargs(for_yaml=for_yaml)
        if _service_host_env in os.environ:
            out['host'] = os.environ[_service_host_env]
            out['address'] = out['address'].replace(self.host.rstrip('/'),
                                                    out['host'].rstrip('/'))
        else:
            out['host'] = self.host
        return out

    def _close(self, *args, **kwargs):
        r"""Close the connection."""
        self._is_open = False
        r = requests.get(
            self.address + '/remove',
            params=self.params,
            cookies=self.cookies)
        r.raise_for_status()
        
    def _send(self, payload):
        r = requests.post(
            self.address,
            data=payload,
            params=self.params,
            cookies=self.cookies)
        r.raise_for_status()
        return True

    def _recv(self, **kwargs):
        r = requests.get(
            self.address,
            params=self.params,
            cookies=self.cookies)
        r.raise_for_status()
        msg = r.content
        if msg == b'':
            raise NoMessages("No messages queued on the server.")
        return (True, msg)

    @property
    def n_msg_recv(self):
        r"""int: The number of incoming messages in the connection."""
        try:
            r = requests.get(
                self.address + '/size',
                params=self.params,
                cookies=self.cookies)
            r.raise_for_status()
            return int(r.content)
        except requests.exceptions.RequestException:  # pragma: debug
            return 0

    @property
    def n_msg_send(self):
        r"""int: The number of outgoing messages in the connection."""
        return self.n_msg_recv

    def purge(self):
        r"""Purge all messages from the comm."""
        r = requests.get(
            self.address + '/purge',
            params=self.params,
            cookies=self.cookies)
        r.raise_for_status()
        self._n_sent = 0
        self._n_recv = 0
        self._last_send = None
        self._last_recv = None



import os


def get_testing_options():
    r"""Get testing parameters for this example."""
    return dict(
        input_files='input.txt',
        expected_output_files=[
            os.path.join('Output', f'expected{i}.txt') for i in range(2)],
        output_dir=None,
        output_files=[f'output{i}.txt' for i in range(2)])



import os
from yggdrasil import tools
from yggdrasil.runner import run
tyk_dir = 'tyk1'

try:
    # Change to the solution directory
    old_dir = os.getcwd()
    if not old_dir.endswith(tyk_dir):
        os.chdir(os.path.join('solutions', tyk_dir))
    if not os.path.isdir('output'):
        os.mkdir('output')

    # Part 1: YAML
    tools.display_source('yamls/weather.yml', number_lines=True)

    # Part 2: Run
    run(['yamls/weather.yml'], production_run=True)

    # Part 3: Open ended...

finally:
    os.chdir(old_dir)



import numpy as np
from yggdrasil import units


def temp(intensity):
    r"""Compute the temperature based on the intensity of light.

    Args:
        intensity (float): Intensity of light in ergs cm^-2 s^-1.

    """
    if not units.has_units(intensity):
        intensity = units.add_units(intensity, 'erg/(cm**2*s)')
    sigma = units.add_units(5.67e-8, 'W m^-2 K^-4')
    T = (np.pi * intensity / sigma)**0.25
    T.convert_to_cgs()
    return T



import pytest
from tests.serialize import TestSerializeBase as base_class


class TestMatSerialize(base_class):
    r"""Test class for TestMatSerialize class."""

    @pytest.fixture(scope="class", autouse=True)
    def serializer(self):
        r"""str: Serializer being tested."""
        return "mat"

    def test_serialize_errors(self, instance):
        r"""Test serialize errors."""
        with pytest.raises(TypeError):
            instance.serialize(['blah', 'blah'])



import pytest
from tests import TestComponentBase as base_class


class TestFilterBase(base_class):
    r"""Test for FilterBase communication flass."""

    _component_type = 'filter'

    @pytest.fixture(scope="class", autouse=True)
    def component_subtype(self, filter):
        r"""Subtype of component being tested."""
        return filter

    @pytest.fixture(scope="class", autouse=True)
    def filter(self, request):
        r"""str: Filter being tested."""
        return request.param

    @pytest.fixture
    def instance_kwargs(self, testing_options):
        r"""Keyword arguments for a new instance of the tested class."""
        out = {}
        if testing_options:
            out = dict(testing_options[0].get('kwargs', {}))
        return out

    def test_filter(self, python_class, testing_options):
        r"""Test filter."""
        for x in testing_options:
            inst = python_class(**x.get('kwargs', {}))
            for msg in x.get('pass', []):
                assert(inst(msg) is True)
            for msg in x.get('fail', []):
                assert(inst(msg) is False)
            for msg, err in x.get('error', []):
                with pytest.raises(err):
                    inst(msg)



import pytest
from tests.drivers.test_BuildModelDriver import (
    TestBuildModelDriver as base_class)
import os
import re
import pprint
import tempfile
from yggdrasil import platform
from yggdrasil.drivers.CMakeModelDriver import (
    CMakeModelDriver, CMakeConfigure, CMakeBuilder)
from yggdrasil.drivers.CModelDriver import GCCCompiler
from yggdrasil.drivers.CPPModelDriver import CPPModelDriver


@pytest.mark.related_language('cmake')
def test_CMakeConfigure(scripts):
    r"""Test CMakeConfigure."""
    src = scripts['c'][0]
    sourcedir = os.path.dirname(src)
    builddir = sourcedir
    # Test get_output_file
    out = CMakeConfigure.get_output_file(src, dont_build=True)
    assert(out == builddir)
    out = CMakeConfigure.get_output_file(src, dont_build=True,
                                         builddir='.', working_dir=sourcedir)
    assert(out == builddir)
    # Test get_flags
    out_A = CMakeConfigure.get_flags(dont_link=True)
    out_B = CMakeConfigure.get_flags(dont_link=True, outfile='.')
    assert(out_A == out_B)


@pytest.mark.related_language('cmake')
def test_CMakeBuilder(scripts):
    r"""Test CMakeBuilder."""
    src = scripts['c'][0]
    target = os.path.splitext(os.path.basename(src))[0]
    builddir = os.path.dirname(src)
    obj = os.path.splitext(src)[0] + '.obj'
    out = os.path.splitext(src)[0]
    if platform._is_win:  # pragma: windows
        out += '.exe'
    # Test get_output_file
    assert(CMakeBuilder.get_output_file(obj) == out)
    assert(CMakeBuilder.get_output_file(obj, target='clean') == 'clean')
    assert(CMakeBuilder.get_output_file(builddir, target=target) == out)
    with pytest.raises(RuntimeError):
        CMakeBuilder.get_output_file(builddir)
    # Test get_flags
    out_A = CMakeBuilder.get_flags(target=target, working_dir=builddir)
    out_B = CMakeBuilder.get_flags(target=target,
                                   outfile=os.path.join('.', os.path.basename(out)))
    assert(out_A == out_B)


@pytest.mark.language('cmake')
def test_create_include():
    r"""Test create_include."""
    target = 'target'
    tempdir = tempfile.gettempdir()
    fname_dll = os.path.join(tempdir, 'test.dll')
    fname_lib = os.path.join(tempdir, 'test.lib')
    for fname in [fname_dll, fname_lib]:
        with open(fname, 'w') as fd:
            fd.write('')
        assert(os.path.isfile(fname))
    testlist = [(['-DYGG'], [], ['ADD_DEFINITIONS(-DYGG)']),
                (['-Wall'], [], ['ADD_DEFINITIONS(-Wall)']),
                (['/nologo'], [], ['ADD_DEFINITIONS(/nologo)']),
                (['-Iinclude_dir'], [], ['INCLUDE_DIRECTORIES(include_dir)']),
                ([], ['-lm'], ['TARGET_LINK_LIBRARIES(%s -lm)' % target]),
                ([], ['-Llib_dir'], ['LINK_DIRECTORIES(lib_dir)']),
                ([], ['/LIBPATH:"lib_dir"'], ['LINK_DIRECTORIES(lib_dir)']),
                ([], ['m'], ['TARGET_LINK_LIBRARIES(%s m)' % target])]
    if CMakeConfigure.add_libraries:  # pragma: debug
        testlist += [([], [fname_dll], ['ADD_LIBRARY(test SHARED IMPORTED)']),
                     ([], [fname_lib], ['ADD_LIBRARY(test STATIC IMPORTED)'])]
    else:
        tempdir_cp = tempdir
        if platform._is_win:  # pragma: windows
            tempdir_cp = tempdir.replace('\\', re.escape('\\'))
        testlist += [([], [fname_dll], [('FIND_LIBRARY(TEST_LIBRARY NAMES %s '
                                         'test HINTS %s)')
                                        % (os.path.basename(fname_dll), tempdir_cp)]),
                     ([], [fname_lib], [('FIND_LIBRARY(TEST_LIBRARY NAMES %s '
                                         'test HINTS %s)')
                                        % (os.path.basename(fname_lib), tempdir_cp)])]
    from yggdrasil.drivers.CModelDriver import CModelDriver
    CModelDriver.compile_dependencies()
    CMakeModelDriver.compile_dependencies()
    kws = {'compiler': CModelDriver.get_tool('compiler'),
           'linker': CModelDriver.get_tool('linker')}
    for c, l, lines in testlist:
        out = CMakeConfigure.create_include(None, target, compiler_flags=c,
                                            linker_flags=l, verbose=True,
                                            **kws)
        for x in lines:
            try:
                assert(x in out)
            except AssertionError:  # pragma: debug
                print("Could not find '%s':" % x)
                pprint.pprint(out)
                raise
    for fname in [fname_dll, fname_lib]:
        os.remove(fname)
    with pytest.raises(ValueError):
        CMakeConfigure.create_include(
            None, target, compiler_flags=['invalid'], **kws)
    with pytest.raises(ValueError):
        CMakeConfigure.create_include(
            None, target, linker_flags=['-invalid'], **kws)
    with pytest.raises(ValueError):
        CMakeConfigure.create_include(
            None, target, linker_flags=['/invalid'], **kws)


@pytest.mark.absent_language('cmake')
def test_CMakeModelDriver_no_C_library(scripts):  # pragma: windows
    r"""Test CMakeModelDriver error when C library not installed."""
    with pytest.raises(RuntimeError):
        CMakeModelDriver('test', scripts['cmake'])


@pytest.mark.language('cmake')
def test_CMakeModelDriver_error_cmake(scripts):
    r"""Test CMakeModelDriver error for invalid cmake args."""
    makedir, target = os.path.split(scripts['cmake'])
    with pytest.raises(RuntimeError):
        CMakeModelDriver('test', target,
                         sourcedir=makedir, compiler_flags='-P',
                         target_language='c')


@pytest.mark.language('cmake')
def test_CMakeModelDriver_error_notarget(scripts):
    r"""Test CMakeModelDriver error for invalid target."""
    makedir, target = os.path.split(scripts['cmake'])
    with pytest.raises(RuntimeError):
        CMakeModelDriver('test', 'invalid',
                         sourcedir=makedir, target_language='c')


@pytest.mark.language('cmake')
def test_CMakeModelDriver_error_nofile():
    r"""Test CMakeModelDriver error for missing CMakeLists.txt."""
    with pytest.raises(RuntimeError):
        CMakeModelDriver('test', 'invalid',
                         target_language='c')


class TestCMakeModelDriver(base_class):
    r"""Test runner for CMakeModelDriver."""

    @pytest.fixture(scope="class")
    def language(self):
        r"""str: Language being tested."""
        return 'cmake'

    @pytest.fixture
    def builddir(self, sourcedir):
        r"""Directory that build will occur in."""
        return os.path.join(sourcedir, 'build')

    @pytest.fixture(autouse=True)
    def dont_verify_fds(self, verify_count_fds, disable_verify_count_fds):
        r"""Turn off verification, fds linger on windows."""
        yield
    
    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, sourcedir,
                        polling_interval, namespace, source):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(testing_options.get('kwargs', {}),
                    yml={'working_dir': sourcedir},
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace, env_compiler='CXX',
                    env_compiler_flags='CXXFLAGS')

    @pytest.mark.skipif(not platform._is_win, reason="Windows only.")
    @pytest.mark.skipif(not GCCCompiler.is_installed(),
                        reason="GNU compiler not installed.")
    def test_run_model_gcc(self, run_model_instance):
        r"""Test compiling/running test model with gcc."""
        run_model_instance(target_compiler='gcc')

    def test_sbdir(self, instance, sourcedir, builddir):
        r"""Test that source/build directories set correctly."""
        assert(instance.sourcedir == sourcedir)
        assert(instance.builddir == builddir)

    def test_write_wrappers(self, instance):
        r"""Test write_wrappers method with verbosity and existing
        include file."""
        try:
            instance.overwrite = False
            instance.write_wrappers(verbose=False)
            instance.write_wrappers(verbose=True)
            instance.overwrite = True
            instance.write_wrappers(verbose=True)
        finally:
            instance.overwrite = True


class TestCMakeModelDriver_wd(TestCMakeModelDriver):
    r"""Test runner for CMakeModelDriver with working directory."""
    
    @pytest.fixture
    def instance_kwargs(self, testing_options, timeout, sourcedir,
                        polling_interval, namespace, source):
        r"""Keyword arguments for a new instance of the tested class."""
        return dict(testing_options.get('kwargs', {}),
                    yml={'working_dir': sourcedir},
                    timeout=timeout, sleeptime=polling_interval,
                    namespace=namespace, env_compiler='CXX',
                    env_compiler_flags='CXXFLAGS',
                    sourcedir='.', builddir='build',
                    compiler_flags=['-Wdev'], skip_compiler=True)
    
    # Disable instance args?

    def test_call_compiler(self, python_class, instance):
        r"""Test call_compiler without full path."""
        # instance.cleanup()
        CPPModelDriver.compile_dependencies()
        python_class.call_compiler(instance.source_files,
                                   builddir='build',
                                   working_dir=instance.working_dir,
                                   dont_build=True)
        out = instance.model_file
        if platform._is_win:
            out = os.path.join(os.path.dirname(out),
                               'Debug',
                               os.path.basename(out))
        compiler = CPPModelDriver.get_tool('compiler')
        python_class.call_compiler(instance.source_files,
                                   out=out,
                                   builddir='build',
                                   working_dir=instance.working_dir,
                                   overwrite=True,
                                   target_compiler=compiler.toolname,
                                   target_linker=compiler.linker().toolname)



import pytest
import copy
import numpy as np


@pytest.fixture(scope="session")
def container_values():
    r"""list: Value for container tests"""
    return [np.float32(1),
            b'hello', u'hello',
            {'nested': np.int64(2)},
            [np.complex128(4), np.uint8(0)]]


@pytest.fixture(scope="session")
def container_values_count(container_values):
    r"""int: Number of container values."""
    return len(container_values)


@pytest.fixture(scope="session")
def container_definitions():
    r"""list: Definitions for container types."""
    return [{'type': 'float',
             'precision': 32,
             'units': ''},
            {'type': 'bytes',
             'precision': 40,
             'units': ''},
            {'type': 'unicode',
             'precision': 40,
             'units': ''},
            {'type': 'object',
             'properties': {'nested': {'type': 'int',
                                       'precision': 64,
                                       'units': ''}}},
            {'type': 'array',
             'items': [{'type': 'complex',
                        'precision': 128,
                        'units': ''},
                       {'type': 'uint',
                        'precision': 8,
                        'units': ''}]}]


@pytest.fixture(scope="session")
def container_typedefs(container_definitions):
    r"""list: Typedefs for container types."""
    out = []
    for v in container_definitions:
        itypedef = {'type': v['type']}
        if v['type'] == 'object':
            itypedef['properties'] = copy.deepcopy(v['properties'])
        elif v['type'] == 'array':
            itypedef['items'] = copy.deepcopy(v['items'])
        out.append(itypedef)
    return out


@pytest.fixture(scope="session")
def ply_test_value():
    r"""Complex example Ply dictionary."""
    vcoords = np.array([[0, 0, 0, 0, 1, 1, 1, 1],
                        [0, 0, 1, 1, 0, 0, 1, 1],
                        [0, 1, 1, 0, 0, 1, 1, 0]], 'float32').T
    vcolors = np.array([[255, 255, 255, 255, 0, 0, 0, 0],
                        [0, 0, 0, 0, 0, 0, 0, 0],
                        [0, 0, 0, 0, 255, 255, 255, 255]], 'uint8').T
    eindexs = np.array([[0, 1, 2, 3, 2],
                        [1, 2, 3, 0, 0]], 'int32')
    ecolors = np.array([[255, 255, 255, 255, 0],
                        [255, 255, 255, 255, 0],
                        [255, 255, 255, 255, 0]], 'uint8')
    out = {'material': 'fake_material', 'vertices': [], 'edges': [],
           'faces': [{'vertex_index': [0, 1, 2]},
                     {'vertex_index': [0, 2, 3]},
                     {'vertex_index': [7, 6, 5, 4]},
                     {'vertex_index': [0, 4, 5, 1]},
                     {'vertex_index': [1, 5, 6, 2]},
                     {'vertex_index': [2, 6, 7, 3]},
                     {'vertex_index': [3, 7, 4, 0]}]}
    for i in range(len(vcoords)):
        ivert = {}
        for j, k in enumerate('xyz'):
            ivert[k] = vcoords[i, j]
        for j, k in enumerate(['red', 'green', 'blue']):
            ivert[k] = vcolors[i, j]
        out['vertices'].append(ivert)
    for i in range(len(eindexs)):
        iedge = {}
        for j, k in enumerate(['vertex1', 'vertex2']):
            iedge[k] = eindexs[i, j]
        for j, k in enumerate(['red', 'green', 'blue']):
            iedge[k] = ecolors[i, j]
        out['edges'].append(iedge)
    for f in out['faces']:
        f['vertex_index'] = [np.int32(x) for x in f['vertex_index']]
    return out


@pytest.fixture(scope="session")
def ply_test_value_simple(ply_test_value):
    r"""Simple example Ply dictionary."""
    return {'vertices': copy.deepcopy(ply_test_value['vertices']),
            'faces': [{'vertex_index': [0, 1, 2]},
                      {'vertex_index': [0, 2, 3]}]}


@pytest.fixture(scope="session")
def ply_test_value_int64(ply_test_value):
    r"""Version of example Ply dictionary using 64bit integers."""
    out = copy.deepcopy(ply_test_value)
    for f in out['faces']:
        f['vertex_index'] = [np.int64(x) for x in f['vertex_index']]
    return out


@pytest.fixture(scope="session")
def obj_test_value(ply_test_value):
    r"""Complex example Obj dictionary."""
    out = {'vertices': [], 'faces': [], 'lines': [],
           'params': [{'u': 0.0, 'v': 0.0, 'w': 0.5},
                      {'u': 0.0, 'v': 0.0}],
           'normals': [{'i': 0.0, 'j': 0.0, 'k': 0.0},
                       {'i': 1.0, 'j': 1.0, 'k': 1.0}],
           'texcoords': [{'u': 0.0, 'v': 0.0, 'w': 0.5},
                         {'u': 1.0, 'v': 0.5},
                         {'u': 1.0}],
           'points': [[0, 2]],
           'curves': [{'starting_param': 0.0, 'ending_param': 1.0,
                       'vertex_indices': [0, 1]}],
           'curve2Ds': [[0, 1]],
           'surfaces': [{
               'starting_param_u': 0.0, 'ending_param_u': 1.0,
               'starting_param_v': 0.0, 'ending_param_v': 1.0,
               'vertex_indices': [{'vertex_index': 0,
                                   'texcoord_index': 0,
                                   'normal_index': 0},
                                  {'vertex_index': 1,
                                   'texcoord_index': 1,
                                   'normal_index': 1}]}]}
    out['material'] = ply_test_value['material']
    out['vertices'] = copy.deepcopy(ply_test_value['vertices'])
    out['vertices'][0]['w'] = 0.5
    for f in ply_test_value['faces']:
        new = [{'vertex_index': x, 'texcoord_index': 0, 'normal_index': 0}
               for x in f['vertex_index']]
        out['faces'].append(new)
    for e in ply_test_value['edges']:
        new = [{'vertex_index': e['vertex%d' % x]} for x in [1, 2]]
        out['lines'].append(new)
    return out


@pytest.fixture(scope="session")
def obj_test_value_simple(obj_test_value):
    r"""Simple example Obj dictionary."""
    out = {'vertices': copy.deepcopy(obj_test_value['vertices']),
           'normals': copy.deepcopy(obj_test_value['normals']),
           'texcoords': copy.deepcopy(obj_test_value['texcoords']),
           'faces': [[{'vertex_index': 0, 'normal_index': 0},
                      {'vertex_index': 1, 'normal_index': 0},
                      {'vertex_index': 2, 'normal_index': 0}],
                     [{'vertex_index': 0, 'normal_index': 1},
                      {'vertex_index': 2, 'normal_index': 1},
                      {'vertex_index': 3, 'normal_index': 1}]]}
    for f in out['faces']:
        for fv in f:
            fv['texcoord_index'] = 0
    for v in out['vertices']:
        v.pop('w', None)
    for t in out['texcoords']:
        t.pop('w', None)
        t.setdefault('v', 0.0)
    return out



import pytest
import numpy as np
from tests.serialize import TestSerializeBase as base_class
from yggdrasil.serialize import DefaultSerialize


class TestFunctionalSerialize(base_class):
    r"""Test class for FunctionalSerialize."""

    @pytest.fixture(scope="class", autouse=True)
    def serializer(self):
        r"""str: Serializer being tested."""
        return "functional"

    @pytest.fixture(scope="class", autouse=True,
                    params=[{'as_format': True}])
    def options(self, request):
        r"""Arguments that should be provided when getting testing options."""
        return request.param

    @pytest.fixture(scope="class")
    def testing_options(self, func_serialize, func_deserialize):
        r"""Testing options."""
        out = {'kwargs': {'func_serialize': func_serialize,
                          'func_deserialize': func_deserialize},
               'empty': b'',
               'objects': [['one', np.int32(1), 1.0],
                           ['two', np.int32(2), 1.0]],
               'extra_kwargs': {},
               'typedef': {'type': 'bytes'},
               'dtype': None,
               'is_user_defined': True}
        return out
    
    @pytest.fixture(scope="class")
    def func_serialize(self):
        r"""Method that serializes using repr."""
        def func_serialize_w(args):
            return repr(args).encode("utf-8")
        return func_serialize_w

    @pytest.fixture(scope="class")
    def func_deserialize(self):
        r"""Method that deserializes using eval."""
        def func_deserialize_w(args):
            if len(args) == 0:
                return b''
            x = eval(args.decode("utf-8"))
            return x
        return func_deserialize_w

    def test_serialize_sinfo(self, instance, testing_options):
        r"""Test serialize/deserialize with serializer info."""
        with pytest.raises(RuntimeError):
            instance.serialize(testing_options['objects'][0],
                               add_serializer_info=True)
    

class FakeSerializer(DefaultSerialize.DefaultSerialize):
    r"""Fake serializer that mocks user defined serialization/deserialization
    routines."""

    _dont_register = True

    def func_serialize(self, args):  # pragma: no cover
        r"""Method that serializes using repr."""
        return repr(args).encode("utf-8")

    def func_deserialize(self, args):  # pragma: no cover
        r"""Method that deserializes using eval."""
        if len(args) == 0:
            return []
        x = eval(args.decode("utf-8"))
        return x


class TestFunctionalSerialize_class(TestFunctionalSerialize):
    r"""Test class for FunctionalSerialize class with classes."""

    @pytest.fixture(scope="class")
    def testing_options(self, func_serialize, func_deserialize):
        r"""Testing options."""
        temp_seri = FakeSerializer()
        assert(issubclass(temp_seri.__class__,
                          DefaultSerialize.DefaultSerialize))
        out = {'kwargs': {'func_serialize': temp_seri,
                          'func_deserialize': temp_seri,
                          'encoded_datatype': {'type': 'bytes'}},
               'empty': b'',
               'objects': [['one', np.int32(1), 1.0],
                           ['two', np.int32(2), 1.0]],
               'extra_kwargs': {},
               'typedef': {'type': 'bytes'},
               'dtype': None,
               'is_user_defined': True}
        return out
        

class TestFunctionalSerialize_error(TestFunctionalSerialize):
    r"""Test class for FunctionalSerialize class with incorrect functions."""

    test_serialize_no_metadata = None
    test_serialize_header = None
    test_serialize_sinfo = None
    test_field_specs = None

    @pytest.fixture(scope="class")
    def func_serialize(self):
        r"""Method that serializes using repr."""
        def func_serialize_w(args):
            return args
        return func_serialize_w

    def test_serialize(self, instance):
        r"""Test serialize with function that dosn't return correct type."""
        with pytest.raises(TypeError):
            instance.serialize((1,))



import pytest
import copy
from yggdrasil import constants, schema
from yggdrasil.components import import_component
from yggdrasil.serialize import SerializeBase
from tests import TestComponentBase as base_class


def test_demote_string():
    r"""Test format str creation of typedef."""
    x = SerializeBase.SerializeBase(format_str='%s')
    assert(x.typedef == {'type': 'array',
                         'items': [{'type': 'bytes'}]})


_seritypes = sorted([x for x in schema.get_schema()['serializer'].subtypes
                     if x not in ['default', 'table', 'pandas', 'map',
                                  'functional', 'mat', 'pickle']])


class TestSerializeBase(base_class):
    r"""Test class for SerializeBase class."""

    _component_type = 'serializer'
    parametrize_serializer = _seritypes
    
    @pytest.fixture(scope="class", autouse=True)
    def component_subtype(self, serializer):
        r"""Subtype of component being tested."""
        return serializer

    @pytest.fixture(scope="class", autouse=True)
    def serializer(self, request):
        r"""str: Serializer being tested."""
        return request.param

    @pytest.fixture(scope="class")
    def empty_msg(self):
        r"""Empty message."""
        return b''

    @pytest.fixture
    def header_info(self):
        r"""Header information."""
        return dict(arg1='1', arg2='two')

    @pytest.fixture(scope="class")
    def empty_head(self):
        def empty_head_w(msg):
            r"""dict: Empty header for message only contains the size."""
            out = dict(size=len(msg), incomplete=False)
            if msg == constants.YGG_MSG_EOF:  # pragma: debug
                out['eof'] = True
            return out
        return empty_head_w

    @pytest.fixture(scope="class")
    def map_sent2recv(self, nested_approx):
        r"""Factory for method to convert sent messages to received."""
        def wrapped_map_sent2recv(obj):
            return nested_approx(obj)
        return wrapped_map_sent2recv

    def test_field_specs(self, instance, testing_options, nested_approx):
        r"""Test field specifiers."""
        assert(instance.numpy_dtype == testing_options['dtype'])
        assert(instance.extra_kwargs
               == nested_approx(testing_options['extra_kwargs']))
        assert(instance.typedef
               == nested_approx(testing_options['typedef']))
        if isinstance(instance.typedef.get('items', []), dict):
            with pytest.raises(Exception):
                instance.get_field_names()
            with pytest.raises(Exception):
                instance.get_field_units()
        else:
            assert(instance.get_field_names()
                   == testing_options.get('field_names', None))
            assert(instance.get_field_units()
                   == testing_options.get('field_units', None))

    def test_concatenation(self, instance, testing_options):
        r"""Test message concatenation."""
        for x, y in testing_options.get('concatenate', []):
            assert(instance.concatenate(x) == y)
        
    def test_serialize(self, instance, map_sent2recv, testing_options,
                       component_subtype, class_name,
                       unyts_equality_patch):
        r"""Test serialize/deserialize."""
        for iobj in testing_options['objects']:
            # if (class_name == 'SerializeBase'):
            #     with pytest.raises(NotImplementedError):
            #         instance.serialize(iobj)
            #     with pytest.raises(NotImplementedError):
            #         instance.serialize(b'test')
            #     break
            msg = instance.serialize(iobj)
            iout, ihead = instance.deserialize(msg)
            print(iout, map_sent2recv(iobj))
            assert(iout == map_sent2recv(iobj))
            # assert(ihead == empty_head(msg))
        if ((('contents' in testing_options)
             and (class_name not in ['SerializeBase', 'DefaultSerialize']))):
            instance.deserialize(testing_options['contents'])

    def test_serialize_no_metadata(self, instance, testing_options):
        r"""Test serializing without metadata."""
        instance.serialize(testing_options['objects'][0],
                           no_metadata=True)

    def test_deserialize_error(self, instance):
        r"""Test error when deserializing message that is not bytes."""
        with pytest.raises(TypeError):
            instance.deserialize(None)
        
    def test_serialize_sinfo(self, instance, testing_options,
                             map_sent2recv, header_info,
                             unyts_equality_patch):
        r"""Test serialize/deserialize with serializer info."""
        hout = copy.deepcopy(header_info)
        temp_seri = import_component(
            'serializer', instance.serializer_info['seritype'])()
        for iobj in testing_options['objects']:
            msg = instance.serialize(iobj, header_kwargs=header_info,
                                     add_serializer_info=True)
            hout.update(instance.serializer_info)
            hout['datatype'] = instance.datatype.encode_type(
                iobj, typedef=instance.typedef)
            iout, ihead = instance.deserialize(msg)
            hout.update(size=ihead['size'], id=ihead['id'],
                        incomplete=False)
            assert(iout == map_sent2recv(iobj))
            assert(ihead == hout)
            # Use info to reconstruct serializer
            iout, ihead = temp_seri.deserialize(msg)
            assert(iout == map_sent2recv(iobj))
            assert(ihead == hout)
            new_seri = import_component('serializer',
                                        ihead.pop('seritype', None))(**ihead)
            iout, ihead = new_seri.deserialize(msg)
            assert(iout == map_sent2recv(iobj))
            assert(ihead == hout)
            
    def test_serialize_header(self, instance, testing_options, header_info,
                              map_sent2recv, unyts_equality_patch):
        r"""Test serialize/deserialize with header."""
        for iobj in testing_options['objects']:
            msg = instance.serialize(iobj, header_kwargs=header_info)
            iout, ihead = instance.deserialize(msg)
            assert(iout == map_sent2recv(iobj))
            # assert(ihead == header_info)
        
    def test_serialize_eof(self, instance):
        r"""Test serialize/deserialize EOF."""
        iobj = constants.YGG_MSG_EOF
        msg = instance.serialize(iobj)
        iout, ihead = instance.deserialize(msg)
        assert(iout == iobj)
        # assert(ihead == empty_head(msg))
        
    def test_serialize_eof_header(self, instance, header_info):
        r"""Test serialize/deserialize EOF with header."""
        iobj = constants.YGG_MSG_EOF
        msg = instance.serialize(iobj, header_kwargs=header_info)
        iout, ihead = instance.deserialize(msg)
        assert(iout == iobj)
        # assert(ihead == empty_head(msg))
        
    def test_deserialize_empty(self, instance, empty_msg, empty_head,
                               testing_options, map_sent2recv):
        r"""Test call for empty string."""
        out, head = instance.deserialize(empty_msg)
        assert(out == map_sent2recv(testing_options['empty']))
        assert(head == empty_head(empty_msg))
